@@ -6,13 +6,14 @@
 /*                 Christophe Tribes           - Polytechnique Montreal            */
 /*                                                                                 */
 /*  The copyright of NOMAD - version 4.0.0 is owned by                             */
+/*                 Charles Audet               - Polytechnique Montreal            */
 /*                 Sebastien Le Digabel        - Polytechnique Montreal            */
 /*                 Viviane Rochon Montplaisir  - Polytechnique Montreal            */
 /*                 Christophe Tribes           - Polytechnique Montreal            */
 /*                                                                                 */
-/*  NOMAD v4 has been funded by Rio Tinto, Hydro-Québec, NSERC (Natural Science    */
-/*  and Engineering Research Council of Canada), INOVEE (Innovation en Energie     */
-/*  Electrique and IVADO (The Institute for Data Valorization)                     */
+/*  NOMAD v4 has been funded by Rio Tinto, Hydro-Québec, NSERC (Natural            */
+/*  Sciences and Engineering Research Council of Canada), InnovÉÉ (Innovation      */
+/*  en Énergie Électrique) and IVADO (The Institute for Data Valorization)         */
 /*                                                                                 */
 /*  NOMAD v3 was created and developed by Charles Audet, Sebastien Le Digabel,     */
 /*  Christophe Tribes and Viviane Rochon Montplaisir and was funded by AFOSR       */
@@ -26,8 +27,6 @@
 /*    Polytechnique Montreal - GERAD                                               */
 /*    C.P. 6079, Succ. Centre-ville, Montreal (Quebec) H3C 3A7 Canada              */
 /*    e-mail: nomad@gerad.ca                                                       */
-/*    phone : 1-514-340-6053 #6928                                                 */
-/*    fax   : 1-514-340-5665                                                       */
 /*                                                                                 */
 /*  This program is free software: you can redistribute it and/or modify it        */
 /*  under the terms of the GNU Lesser General Public License as published by       */
@@ -54,13 +53,11 @@
 #ifndef __NOMAD400_CACHEBASE__
 #define __NOMAD400_CACHEBASE__
 
-#ifdef _OPENMP
-#include <omp.h>
-#endif // _OPENMP
-
 #include <atomic>       // For atomic
 #include <vector>
 
+#include "../config.hpp"
+
 #include "../Eval/EvalPoint.hpp"
 #include "../Param/CacheParameters.hpp"
 
@@ -122,18 +119,27 @@ protected:
      */
     size_t _n;
 
+    /// Signals that the cache should stop waiting for points to be evaluated.
+    /**
+     * For methods that allow waiting for a point to be evaluated:
+     * If the whole optimization is done, stop waitings for points
+     * that will never be evaluated.
+     */
+    bool _stopWaiting;
+
     /*---------*/
     /* Methods */
     /*---------*/
 
     /// Protected constructor.
     /**
-     * Called only by derived object to instanciate the singleton
+     * Called only by derived object to instantiate the singleton
      \param cacheParams The cache parameters -- \b IN.
      */
     explicit CacheBase(const std::shared_ptr<CacheParameters>& cacheParams)
       : _cacheParams (cacheParams),
-        _n(0)
+        _n(0),
+        _stopWaiting(false)
     {
         init();
     }
@@ -154,7 +160,7 @@ public:
     {
         if (nullptr == _single)
         {
-            std::string err = "Cannot get instance. A non-virtual object derived from CacheBase must be instanciated first. For example, call CacheSet::setInstance() ONCE before calling CacheBase::getInstance()" ;
+            std::string err = "Cannot get instance. A non-virtual object derived from CacheBase must be instantiated first. For example, call CacheSet::setInstance() ONCE before calling CacheBase::getInstance()" ;
             throw Exception(__FILE__, __LINE__, err);
         }
 
@@ -178,6 +184,8 @@ public:
 
     void setMaxSize(const size_t maxSize) { _maxSize = maxSize; }
 
+    void setStopWaiting(const bool stopWaiting) { _stopWaiting = stopWaiting; }
+
     /*---------------*/
     /* Other methods */
     /*---------------*/
@@ -201,17 +209,26 @@ public:
        each EvalPoint may be processed in place. It is not needed
        to remove it from the cache, process it, and then put it back.
      */
-    virtual void processOnAllPoints(void (*func)(EvalPoint&) __attribute__((unused)))
+    typedef void (*EvalFunc_t)(EvalPoint&);
+    virtual void processOnAllPoints(EvalFunc_t NOMAD_UNUSED(func))
     {
         std::cerr << "Warning: processOnAllPoints is not implemented for this type of cache." << std::endl;
     }
+    virtual void processOnAllPoints(EvalFunc_t NOMAD_UNUSED(func), const int NOMAD_UNUSED(mainThreadNum))
+    {
+        std::cerr << "Warning: processOnAllPoints is not implemented for this type of cache." << std::endl;
+    }
+
+    virtual void deleteSgteOnly(const int mainThreadNum) = 0;
+
+
 
     /// Add a new EvalPoint to the cache.
     /**
      * If insertion worked, the point was not in the cache before. Return true.\n
      * If insertion did not work, the point was in the cache before.
        \c _nbCacheHits is incremented. Return false.
-
+       \note This implementation calls smartInsert().
      \param  evalPoint   The eval point to insert in cache -- \b IN.
      \return             \c true if insertion works and \c false if not.
      */
@@ -223,21 +240,24 @@ public:
 
      \param x           The point to find                       -- \b IN.
      \param evalPoint   The returned eval point that matches x  -- \b IN.
+     \param evalType    If not UNDEFINED, wait for the point to have an evaluation for this evaltype. -- \b IN.
      \return            The number of eval points found.
      */
-    virtual size_t find(const Point & x, EvalPoint &evalPoint) const = 0;
+    virtual size_t find(const Point & x, EvalPoint &evalPoint,
+                        const EvalType evalType = EvalType::UNDEFINED) const = 0;
 
     /// Insert evalPoint in cache.
     /**
+     * evalPoint's tag (mutable) is updated.
      * If insertion worked, the point was not in the cache before. Return true.
      * If insertion did not work, the point was in the cache before.
       _nbCacheHits is incremented.
      * Depending on its EvalStatus, on maxNumberEval, and on the evalType,
        return \c true if it should be evaluated again,
        \c false otherwise.
-     \param evalPoint       The eval point to insert in the cache -- \b IN.
-     \param maxNumberEval   The max number of evals           -- \b IN.
-     \param evalType        Look at the Blackbox or Surrogate eval of the EvalPoint  -- \b IN.
+     \param evalPoint           The eval point to insert in the cache -- \b IN.
+     \param maxNumberEval  The max number of evals           -- \b IN.
+     \param evalType             Look at the Blackbox or Surrogate eval of the EvalPoint  -- \b IN.
      \return                A boolean indicating if we should eval this point.
      */
     virtual bool smartInsert(const EvalPoint &evalPoint,
@@ -248,9 +268,9 @@ public:
     /**
      Get all eval points at point x from the cache (pure virtual function).
 
-     \param x              The point to find in cache             -- \b IN.
+     \param x                             The point to find in cache             -- \b IN.
      \param evalPointList  The list of eval points found in cache -- \b OUT.
-     \return               The number of points found.
+     \return                The number of points found.
      */
     virtual size_t find(const Point x,
                         std::vector<EvalPoint> &evalPointList) const = 0;
@@ -259,46 +279,50 @@ public:
     /// Get all eval points for which comp(refeval) returns true.
     /**
      The comparison function tests if an eval point's eval is inferior to refeval.
-     \param refeval         The point of reference                                      -- \b IN.
-     \param comp            The comparison function                                     -- \b IN.
+     \param refeval                The point of reference                                      -- \b IN.
+     \param comp                       The comparison function                                     -- \b IN.
      \param evalPointList   The list of eval points found in cache that match comp()    -- \b OUT.
-     \param evalType        Look at the blackbox or surrogate eval of the EvalPoint     -- \b IN.
+     \param evalType              Look at the blackbox or surrogate eval of the EvalPoint     -- \b IN.
      \return                The number of points found.
      */
     virtual size_t find(const Eval &refeval,
-                     bool (*comp)(const Eval&, const Eval&),
-                     std::vector<EvalPoint> &evalPointList,
-                     const EvalType& evalType = EvalType::BB) const = 0;
+                        std::function<bool(const Eval&, const Eval&)> comp,
+                        std::vector<EvalPoint> &evalPointList,
+                        const EvalType& evalType = EvalType::BB) const = 0;
 
 
     /// Get best eval points, using comp(). Only the points with eval status EVAL_OK are considered.
     /**
-     \param comp             The comparison function                                    -- \b IN.
-     \param evalPointList    The best eval points that verify comp()==true  in a list   -- \b OUT.
-     \param findFeas         The flag to find feasible points                           -- \b IN.
-     \param hMax             The hmax to detect feasibility                             -- \b IN.
-     \param fixedVariable   Searching for a subproblem defined by this point -- \b IN.
-     \param evalType         Which Eval of the EvalPoint to look at  -- \b IN.
+     \param comp                         The comparison function                                    -- \b IN.
+     \param evalPointList     The best eval points that verify comp()==true  in a list   -- \b OUT.
+     \param findFeas                The flag to find feasible points                           -- \b IN.
+     \param hMax                         The hmax to detect feasibility                             -- \b IN.
+     \param fixedVariable     Searching for a subproblem defined by this point -- \b IN.
+     \param evalType                Which Eval of the EvalPoint to look at  -- \b IN.
+     \param refeval                  The upper bound eval reference to accelerate the search  (can be nullptr)   -- \b IN.
      \return                 The number of eval points found.
      */
-    virtual size_t findBest(bool (*comp)(const Eval&, const Eval&),
-                     std::vector<EvalPoint> &evalPointList,
-                     const bool findFeas,
-                     const Double& hMax,
-                     const Point& fixedVariable,
-                     const EvalType& evalType) const = 0;
+    virtual size_t findBest(std::function<bool(const Eval&, const Eval&)> comp,
+                            std::vector<EvalPoint> &evalPointList,
+                            const bool findFeas,
+                            const Double& hMax,
+                            const Point& fixedVariable,
+                            const EvalType& evalType,
+                            const Eval* refeval) const = 0;
 
 
     /// Find best feasible points in the cache using operator<.
     /**
-     \param evalPointList   The best feasible eval points in a list  -- \b OUT.
+     \param evalPointList  The best feasible eval points in a list  -- \b OUT.
      \param fixedVariable  Searching for a subproblem defined by this point -- \b IN.
-     \param evalType        Which eval (Blackbox or Surrogate) of the EvalPoint to look at  -- \b IN.
+     \param evalType             Which eval (Blackbox or Surrogate) of the EvalPoint to look at  -- \b IN.
+     \param refeval               The upper bound eval reference to accelerate the search  (can be nullptr)   -- \b IN.
      \return                The number of eval points found.
      */
     virtual size_t findBestFeas(std::vector<EvalPoint> &evalPointList,
                             const Point& fixedVariable,
-                            const EvalType& evalType) const = 0;
+                            const EvalType& evalType,
+                            const Eval* refeval) const = 0;
 
     /// Test if cache contains feasible points.
     /**
@@ -308,41 +332,61 @@ public:
 
     /// Find best infeasible points, with h <= hMax, using operator< (pure virtual).
     /**
-     \param evalPointList   The best infeasible eval points in a list  -- \b OUT.
-     \param hMax            Select a point if h <= hMax                -- \b IN.
+     \param evalPointList  The best infeasible eval points in a list  -- \b OUT.
+     \param hMax                      Select a point if h <= hMax                -- \b IN.
      \param fixedVariable  Searching for a subproblem defined by this point -- \b IN.
-     \param evalType        Which eval (Blackbox or Surrogate) of the EvalPoint to look at  -- \b IN.
+     \param evalType             Which eval (Blackbox or Surrogate) of the EvalPoint to look at  -- \b IN.
+     \param refeval                The upper bound eval reference to accelerate the search  (can be nullptr)   -- \b IN.
      \return                The number of eval points found.
      */
     virtual size_t findBestInf(std::vector<EvalPoint> &evalPointList,
                             const Double& hMax,
                             const Point& fixedVariable,
-                            const EvalType& evalType) const = 0;
+                            const EvalType& evalType,
+                            const Eval* refeval) const = 0;
 
     /// Get all eval points within a distance of point X.
     /**
-     \param X                The point of reference                                         -- \b IN.
-     \param distance         The distance to the point of reference                         -- \b IN.
+     \param X                               The point of reference                                         -- \b IN.
+     \param crit                        The criteria function                                 -- \b IN.
      \param evalPointList    The eval points within the prescribed distance of X            -- \b OUT.
      \param maxEvalPoints    The maximum number of points to select                         -- \b IN.
      \return                 The number of eval points found.
      */
-    virtual size_t find(Point X,
-                        Double distance,
+    virtual size_t find(const Point & X,
+                        std::function<bool(const Point&, const EvalPoint &)> crit,
                         std::vector<EvalPoint> &evalPointList,
                         int maxEvalPoints = 0) const = 0;
 
+
+
     /// Find using criteria.
     /**
      All the points for which crit() return \c true are put in evalPointList.
 
-     \param crit             The criteria function                               -- \b IN.
+     \param crit                        The criteria function                               -- \b IN.
      \param evalPointList    The eval points within the prescribed distance of X -- \b OUT.
      \return                 The number of eval points found.
      */
-    virtual size_t find(bool (*crit)(const EvalPoint&),
+    virtual size_t find(std::function<bool(const EvalPoint&)> crit,
                         std::vector<EvalPoint> &evalPointList) const = 0;
 
+
+
+    /// Get all eval points using two custom criterions.
+    /**
+     All the points for which the two crit() functions return \c true are put in evalPointList.
+
+     \param crit1                      The first criteria function                               -- \b IN.
+     \param crit2                      The first criteria function                               -- \b IN.
+     \param evalPointList    The eval points within the prescribed distance of X            -- \b OUT.
+     \return                 The number of eval points found.
+     */
+    virtual size_t find(std::function<bool(const EvalPoint&)> crit1,
+                        std::function<bool(const EvalPoint&)> crit2,
+                        std::vector<EvalPoint> &evalPointList) const = 0;
+
+
     // More find() methods can be added here.
     // Find on multiple points -returning multiple eval points
     // Etc.
@@ -373,7 +417,7 @@ public:
     virtual bool clear() = 0;
 
     /// Clear all sgte evaluations from the cache.
-    virtual void clearSgte() = 0;
+    virtual void clearSgte(const int mainThreadNum) = 0;
 
     /**
      * \brief Purge the cache from elements with higher f.
@@ -392,6 +436,13 @@ public:
      */
     virtual bool write() const = 0;
 
+    /**
+     * \brief Display all points in cache.
+     *
+     * \note Used for debugging.
+     */
+    virtual std::string displayAll() const { return ""; }
+
     /// Read a cache file and load it.
     virtual bool read() = 0;
 
