@@ -6,13 +6,14 @@
 /*                 Christophe Tribes           - Polytechnique Montreal            */
 /*                                                                                 */
 /*  The copyright of NOMAD - version 4.0.0 is owned by                             */
+/*                 Charles Audet               - Polytechnique Montreal            */
 /*                 Sebastien Le Digabel        - Polytechnique Montreal            */
 /*                 Viviane Rochon Montplaisir  - Polytechnique Montreal            */
 /*                 Christophe Tribes           - Polytechnique Montreal            */
 /*                                                                                 */
-/*  NOMAD v4 has been funded by Rio Tinto, Hydro-Québec, NSERC (Natural Science    */
-/*  and Engineering Research Council of Canada), INOVEE (Innovation en Energie     */
-/*  Electrique and IVADO (The Institute for Data Valorization)                     */
+/*  NOMAD v4 has been funded by Rio Tinto, Hydro-Québec, NSERC (Natural            */
+/*  Sciences and Engineering Research Council of Canada), InnovÉÉ (Innovation      */
+/*  en Énergie Électrique) and IVADO (The Institute for Data Valorization)         */
 /*                                                                                 */
 /*  NOMAD v3 was created and developed by Charles Audet, Sebastien Le Digabel,     */
 /*  Christophe Tribes and Viviane Rochon Montplaisir and was funded by AFOSR       */
@@ -26,8 +27,6 @@
 /*    Polytechnique Montreal - GERAD                                               */
 /*    C.P. 6079, Succ. Centre-ville, Montreal (Quebec) H3C 3A7 Canada              */
 /*    e-mail: nomad@gerad.ca                                                       */
-/*    phone : 1-514-340-6053 #6928                                                 */
-/*    fax   : 1-514-340-5665                                                       */
 /*                                                                                 */
 /*  This program is free software: you can redistribute it and/or modify it        */
 /*  under the terms of the GNU Lesser General Public License as published by       */
@@ -57,8 +56,9 @@
 
 #include <functional>   // For std::function
 
+#include "../config.hpp"
+
 #include "../Eval/BBOutput.hpp"
-#include "../Math/Double.hpp"
 #include "../Param/EvalParameters.hpp"
 
 #include "../nomad_nsbegin.hpp"
@@ -88,6 +88,7 @@ enum class EvalStatusType
     EVAL_CONS_H_OVER,       ///< Evaluation was rejected because constraint violation was higher than hMax. May be submitted again.
     EVAL_OK,                ///< Correct evaluation
     EVAL_IN_PROGRESS,       ///< Evaluation in progress
+    EVAL_WAIT,              ///< Evaluation in progress for another instance of the same point: Wait for evaluation to be done.
     EVAL_STATUS_UNDEFINED   ///< Undefined evaluation status
 };
 
@@ -109,9 +110,11 @@ private:
     Double _h;   ///< Value of the constraint violation
     EvalStatusType _evalStatus;  ///> The evaluation status.
     BBOutput _bbOutput;  ///<  The blackbox evaluation output.
+    bool _bbOutputComplete;  ///< All bbo outputs have a valid value for functions (OBJ, PB and EB).
 
+    // This method should not be static. Issue #410.
     static std::function<SuccessType(const Eval* eval1, const Eval* eval2, const Double& hMax)> _computeSuccessType;  ///< The function called to compute success type.
-    
+
     /// The function that computes h from an Eval and a list of blackbox output types.
     /**
      * For a given Eval there can be several ways to compute infeasibility
@@ -119,6 +122,8 @@ private:
      PhaseOneSearch of Mads or after the PhaseOneSearch the computation is
      different. This function sums the infeasibility measure of all constraints
      (see _computeHComponent function).
+     \note This method should not be static. Issue #410.
+     \note This method to be revised with issue #332.
      */
     static std::function<Double(const Eval& eval, const
                                 BBOutputTypeList &bbOutputTypeList)> _computeH;
@@ -129,6 +134,8 @@ private:
      given output type (::BBOutputType). \n
      * A default function is provided in Eval::defaultComputeHComponent that return max(g_i,0)^2. A user can provide a different function using the static Eval::setComputeHComponent function. \n
      * This allows to change the computation of h or relax the constraint bounds ( hMim = 0 -> hMin>0).
+     \note This method should not be static. Issue #410.
+     \note This method to be revised with issue #332.
      */
     static std::function<Double(const BBOutputType &bbOutputType,
                                 size_t index,
@@ -174,6 +181,8 @@ public:
     EvalStatusType getEvalStatus() const { return _evalStatus; }
     void setEvalStatus(const EvalStatusType &evalStatus) { _evalStatus = evalStatus; }
 
+    bool isBBOutputComplete () const { return _bbOutputComplete; }
+
     BBOutput getBBOutput() const { return _bbOutput; }
     void setBBOutput(const BBOutput &bbOutput);
 
@@ -193,7 +202,7 @@ public:
     /*---------------*/
 
     bool toBeRecomputed() const { return _toBeRecomputed; }
-    void toRecompute(bool toBeRecomputed) { _toBeRecomputed = toBeRecomputed; }
+
 
     /// Compute objective function value.
     /**
@@ -223,7 +232,7 @@ public:
      \return              The component of h for constraint referred by index.
      */
     static Double defaultComputeHComponent( const BBOutputType & bbOutputType ,
-                                            size_t index __attribute__((unused)),
+                                            size_t NOMAD_UNUSED(index),
                                             const Double &bbo );
 
 
@@ -247,8 +256,8 @@ public:
     /** Should this point be saved to cache file? Based on the eval status only.
      * These eval statuses are good: EVAL_OK, EVAL_FAILED, EVAL_USER_REJECTED,
      * EVAL_CONS_H_OVER, EVAL_ERROR.
-     * These eval statuses are not good: 
-     * EVAL_NOT_STARTED, EVAL_IN_PROGRESS, EVAL_STATUS_UNDEFINED.
+     * These eval statuses are not good:
+     * EVAL_NOT_STARTED, EVAL_IN_PROGRESS, EVAL_WAIT, EVAL_STATUS_UNDEFINED.
     */
     bool goodForCacheFile() const;
 
@@ -277,10 +286,10 @@ public:
      */
     bool operator<(const Eval& eval) const;
 
-    /// Dominance 
+    /// Dominance
     /**
      Dominace as by definition 12.3 in the Book of Audet and Hare, Derivative-Free and Blackbox Optimization,
-     https://doi.org/10.1007/978-3-319-68913-5 
+     https://doi.org/10.1007/978-3-319-68913-5
      * The feasible point x dominates the feasible point y \n
        when f(x) < f(y).
      * The infeasible point x dominates the infeasible point y \n
@@ -329,7 +338,7 @@ public:
      */
     static SuccessType computeSuccessTypePhaseOne(const Eval* eval1,
                                                   const Eval* eval2,
-                                                  const Double& hMax  __attribute__((unused)));
+                                                  const Double& NOMAD_UNUSED(hMax));
 
     /// Set which function should be used to compute success type
     static void setComputeSuccessTypeFunction(const std::function<SuccessType(
