@@ -6,13 +6,14 @@
 /*                 Christophe Tribes           - Polytechnique Montreal            */
 /*                                                                                 */
 /*  The copyright of NOMAD - version 4.0.0 is owned by                             */
+/*                 Charles Audet               - Polytechnique Montreal            */
 /*                 Sebastien Le Digabel        - Polytechnique Montreal            */
 /*                 Viviane Rochon Montplaisir  - Polytechnique Montreal            */
 /*                 Christophe Tribes           - Polytechnique Montreal            */
 /*                                                                                 */
-/*  NOMAD v4 has been funded by Rio Tinto, Hydro-Québec, NSERC (Natural Science    */
-/*  and Engineering Research Council of Canada), INOVEE (Innovation en Energie     */
-/*  Electrique and IVADO (The Institute for Data Valorization)                     */
+/*  NOMAD v4 has been funded by Rio Tinto, Hydro-Québec, NSERC (Natural            */
+/*  Sciences and Engineering Research Council of Canada), InnovÉÉ (Innovation      */
+/*  en Énergie Électrique) and IVADO (The Institute for Data Valorization)         */
 /*                                                                                 */
 /*  NOMAD v3 was created and developed by Charles Audet, Sebastien Le Digabel,     */
 /*  Christophe Tribes and Viviane Rochon Montplaisir and was funded by AFOSR       */
@@ -26,8 +27,6 @@
 /*    Polytechnique Montreal - GERAD                                               */
 /*    C.P. 6079, Succ. Centre-ville, Montreal (Quebec) H3C 3A7 Canada              */
 /*    e-mail: nomad@gerad.ca                                                       */
-/*    phone : 1-514-340-6053 #6928                                                 */
-/*    fax   : 1-514-340-5665                                                       */
 /*                                                                                 */
 /*  This program is free software: you can redistribute it and/or modify it        */
 /*  under the terms of the GNU Lesser General Public License as published by       */
@@ -54,15 +53,12 @@
 #include <typeindex>
 #include <typeinfo>
 
-#include "../Util/defines.hpp"
 #include "../Math/Double.hpp"
 #include "../Math/Point.hpp"
 #include "../Math/ArrayOfPoint.hpp"
-#include "../Param/Attribute.hpp"
+#include "../Type/ListOfVariableGroup.hpp"
 #include "../Param/AttributeFactory.hpp"
 #include "../Param/ParameterEntries.hpp"
-#include "../Util/ArrayOfString.hpp"
-#include "../Util/fileutils.hpp"
 
 
 #include "../nomad_nsbegin.hpp"
@@ -102,23 +98,37 @@ struct AttributeDefinition
 
 
 /// Exception class for an invalid parameter.
-class InvalidParameter : public NOMAD::Exception
+class InvalidParameter : public Exception
 {
 public:
     /// Constructor.
     InvalidParameter(const std::string& file,
                      int line,
                      const std::string& msg)
-      : NOMAD::Exception(file, line, msg)
+      : Exception(file, line, msg)
     {
-        _typeMsg = "Invalid Parameter";
+        _typeMsg = "Invalid Parameter.";
+    }
+};
+
+/// Exception class for a parameter that has not been checked.
+class ParameterToBeChecked : public Exception
+{
+public:
+    /// Constructor.
+    ParameterToBeChecked(const std::string& file,
+                     int line,
+                     const std::string& msg)
+      : Exception(file, line, msg)
+    {
+        _typeMsg = "Parameter to be checked.";
     }
 };
 
 
 /// Abstract class for the NOMAD parameters.
 /**
- Several types of parameters control the execution of Nomad: RunParameters, PbParameters, CacheParameters, DisplayParameters, EvalParameters and EvaluatorControlParameters. \n
+ Several types of parameters control the execution of Nomad: RunParameters, PbParameters, CacheParameters, DisplayParameters, EvalParameters, EvaluatorControlGlobalParameters and EvaluatorControlParameters. \n
 
  All the parameters to control a NOMAD run can be obtained from a single parameter file (see Parameters::readParamFileAndSetEntries) or set using the attribute name and value (see the templated function Parameters::setAttributeValue). \n
 
@@ -208,7 +218,7 @@ protected:
     /**
      An attribute is defined by some meta data provided in the structure AttributeDefinition. This structure contains a series of string that are translated into Attribute -s- when calling Parameters::registerAttributes. \n
 
-     This definition is provided as an "initializer list" of a vector of several strings: _definition = { {"attribute1_name","attribute1_type","attribute1_defaultValue",....},{"attribute2_name","attribute2_type",....},...} in the header files for specific attribute definition files: cacheAttributesDefinition.hpp, displayAttributesDefinition.hpp, evalAttributesDefinition.hpp, evaluatorControlAttributesDefinition.hpp, pbAttributesDefinition.hpp and runAttributesDefinition.hpp. These files are programmatically created from their equivalent text files (*.txt) by the WriteAttributeDefinition.exe binary. This pre-processing task is automatically performed (by makefile) if an attribute definition is modified in a text file.\n
+     This definition is provided as an "initializer list" of a vector of several strings: _definition = { {"attribute1_name","attribute1_type","attribute1_defaultValue",....},{"attribute2_name","attribute2_type",....},...} in the header files for specific attribute definition files: cacheAttributesDefinition.hpp, displayAttributesDefinition.hpp, evalAttributesDefinition.hpp, evaluatorControlAttributesDefinition.hpp, evaluatorControlGlobalAttributesDefinition.hpp, pbAttributesDefinition.hpp and runAttributesDefinition.hpp. These files are programmatically created from their equivalent text files (*.txt) by the WriteAttributeDefinition.exe binary. This pre-processing task is automatically performed (by makefile) if an attribute definition is modified in a text file.\n
 
      */
     std::vector<AttributeDefinition> _definition;
@@ -285,6 +295,9 @@ private:
     /// Helper for read
     size_t readValuesForArrayOfPoint(const ParameterEntry &pe, Point &point);
 
+    /// Helper for read
+    size_t readValuesForVariableGroup(const ParameterEntry &pe, VariableGroup &vg);
+
 protected:
     /*-------------------*/
     /* registerAttribute   */
@@ -312,7 +325,7 @@ protected:
             throw Exception(__FILE__,__LINE__, err);
         }
 
-        toupper(name);
+        NOMAD::toupper(name);
 
         auto attribute = AttributeFactory{}.Create<T>(name, default_value,
                                                       algoCompatibilityCheck, restartAttribute, uniqueEntry,
@@ -347,7 +360,9 @@ protected:
 
     const std::string& getAttributeType(const std::string& name)
     {
-        return _typeOfAttributes[name];
+        auto namecaps = name;
+        NOMAD::toupper(namecaps);
+        return _typeOfAttributes[namecaps];
     }
 
     SPtrAtt getAttribute(std::string name) const;
@@ -398,7 +413,7 @@ protected:
             {
                 std::string err = "In getAttributeValue<T> the attribute ";
                 err += name + " has not been checked";
-                throw Exception(__FILE__,__LINE__, err);
+                throw ParameterToBeChecked(__FILE__,__LINE__, err);
             }
             return sp->getValue();
         }
@@ -429,12 +444,14 @@ protected:
     template<typename T> const Point&
     getAttributeValueProtected(const std::string &name, type<Point>, bool flagCheckException, bool flagDefault = false) const
     {
-        if (typeid(ArrayOfPoint).name() == _typeOfAttributes.at(name))
+        auto namecaps = name;
+        NOMAD::toupper(namecaps);
+        if (typeid(ArrayOfPoint).name() == _typeOfAttributes.at(namecaps))
         {
             // Special case: Attribute type is an ArrayOfPoint, but user asks to
             // return a Point.
             // Get the ArrayOfPoint and return its first element.
-            const ArrayOfPoint & aop = getSpValue<ArrayOfPoint>(name, flagCheckException, flagDefault);
+            const ArrayOfPoint & aop = getSpValue<ArrayOfPoint>(namecaps, flagCheckException, flagDefault);
             if (aop.size() >= 1)
             {
                 return aop[0];
@@ -448,7 +465,7 @@ protected:
         }
 
         // Default behaviour
-        return getSpValue<T>(name, flagCheckException, flagDefault);
+        return getSpValue<T>(namecaps, flagCheckException, flagDefault);
     }
 
 
@@ -475,7 +492,9 @@ public:
     template<typename T> const T&
     getAttributeValue(const std::string &name, bool flagDefault = false) const
     {
-        return getAttributeValueProtected<T>(name,true,flagDefault);
+        auto namecaps = name;
+        NOMAD::toupper(namecaps);
+        return getAttributeValueProtected<T>(namecaps,true,flagDefault);
     }
 
     /**
@@ -514,10 +533,12 @@ public:
     {
 
         std::string typeTName = typeid(T).name();
+        auto namecaps = name;
+        NOMAD::toupper(namecaps);
         auto att = getAttribute(name);
 
         // Must use map access with "at" (not []) because the function is const
-        if (typeTName != _typeOfAttributes.at(name))
+        if (typeTName != _typeOfAttributes.at(namecaps))
         {
             std::string err = "In isAttributeDefaultValue<T> : the attribute " + name;
             err += " is not of type T = " + typeTName;
@@ -605,6 +626,29 @@ public:
         }
     }
 
+    /**
+     Overload of setSpValue for std::string -> ArrayOfString case.
+     Value is of type std::string, and it might need to be converted to an ArrayOfString for parameter name.
+     */
+    void setSpValue(const std::string& name, std::string value)
+    {
+        if (typeid(ArrayOfString).name() == _typeOfAttributes.at(name))
+        {
+            // Special case: Attribute type is an ArrayOfString, but user sets
+            // a string.
+            // Create an ArrayOfString and set its first element to the
+            // given string value.
+            ArrayOfString aos;
+            aos.add(value);
+
+            setSpValue(name, aos);
+        }
+        else
+        {
+            // Use default behaviour
+            setSpValueDefault(name, value);
+        }
+    }
 
     /**
      Overload of setSpValue for int -> size_t case.
@@ -647,7 +691,9 @@ public:
             throw Exception(__FILE__, __LINE__, err);
         }
 
-        setSpValue(name, value);
+        auto namecaps = name;
+        NOMAD::toupper(namecaps);
+        setSpValue(namecaps, value);
 
         _toBeChecked = true;
     }
