@@ -1,12 +1,16 @@
 // Copyright UC Regents
 
 #include "PyCompat.h"
+#include "../config.hpp"
 #include "../Nomad/nomad.hpp"
+#include "../Cache/CacheBase.hpp"
 #include "../Type/BBOutputType.hpp"
+#include "../Algos/EvcInterface.hpp"
 #include "../Math/RNG.hpp"
 
-#include <stdexcept>
 #include <memory>
+#include <sstream>
+#include <stdexcept>
 #include <vector>
 
 #include <stdint.h>
@@ -26,12 +30,31 @@ namespace SQNomad {
 
 //- helpers ------------------------------------------------------------------
 namespace {
+
 class PyGILRAII {
     PyGILState_STATE m_GILState;
 public:
     PyGILRAII() : m_GILState(PyGILState_Ensure()) {}
     ~PyGILRAII() { PyGILState_Release(m_GILState); }
 };
+
+class PyUnGILRAII {
+    PyThreadState* m_ThreadState;
+public:
+    PyUnGILRAII() { m_ThreadState = PyEval_SaveThread(); }
+    ~PyUnGILRAII() { PyEval_RestoreThread(m_ThreadState); }
+};
+
+class EvcInterfaceReset {
+public:
+    ~EvcInterfaceReset() {
+    // resetting the evaluator is necessary b/c it's implementation is a static
+    // shared_ptr: since the evaluator is user-provided, it may be destroyed after
+    // the user level library was already unloaded
+        EvcInterface::getEvaluatorControl()->setEvaluator(nullptr);
+    }
+};
+
 } // unnamed namespace
 
 
@@ -156,7 +179,7 @@ private:
 public:
     PyCallback(std::shared_ptr<EvalParameters> ep, PyObject* f) :
             Evaluator(ep) {
-        Py_INCREF(f); Py_INCREF(f);
+        Py_INCREF(f);
         m_callback = f;
     }
     PyCallback(const PyCallback& pc) : Evaluator(pc) {
@@ -175,14 +198,14 @@ public:
 
 public:
     std::vector<bool> eval_block(Block& block,
-            __attribute__((unused)) const Double& hMax,
+            const Double& NOMAD_UNUSED(hMax),
             std::vector<bool>& countEval) const override {
 
         const size_t np = block.size();
         std::vector<bool> evalOk(np, false);
         countEval.resize(np, false);
 
-        auto bboutputl = stringToBBOutputTypeList("OBJ");
+        auto bbOutputType = _evalParams->getAttributeValue<NOMAD::BBOutputTypeList>("BB_OUTPUT_TYPE");
 
      // TODO: this code can be called in parallel by NOMAD; need to test whether
      // to place GIL acquisition inside the loop itself where needed, rathern than
@@ -202,18 +225,41 @@ public:
             double* buf = (double*)PyArray_DATA((PyArrayObject*)py_x);
             for (size_t j = 0; j < x->size(); ++j) buf[j] = (*x)[j].todouble();
             PyObject* res = PyObject_CallObject(m_callback, args);
-            evalOk[i] = (bool)res;
+            evalOk[i] = false;
             if (res) {
-                double dres = PyFloat_AsDouble(res);
-                if (!(dres == -1. && PyErr_Occurred())) {
-                    Eval e{};
-                    e.setF(dres);
-                    e.setH(Eval::defaultComputeH(e, bboutputl));
-                    x_ptr->setEval(e, EvalType::BB);
+                std::ostringstream sbbo;
+
+                if (PyTuple_Check(res)) {
+                    evalOk[i] = true;
+                    for (Py_ssize_t resi = 0; resi < PyTuple_GET_SIZE(res); ++resi) {
+                        double dres = PyFloat_AsDouble(PyTuple_GET_ITEM(res, resi));
+                        if (dres == -1. && PyErr_Occurred()) {
+                            evalOk[i] = false;
+                            break;
+                        }
+                        if (resi != 0)
+                            sbbo << " ";
+                        sbbo << dres;
+                    }
+                } else {
+                    double dres = PyFloat_AsDouble(res);
+                    if (!(dres == -1. && PyErr_Occurred())) {
+                        evalOk[i] = true;
+                        sbbo << dres;
+                    }
                 }
+
+                if (evalOk[i]) {
+                    const NOMAD::EvalType& evalType = getEvalType();
+                    x_ptr->setBBO(sbbo.str(), bbOutputType, evalType);
+                }
+
                 Py_DECREF(res);
-            } else
+            }
+
+            if (!evalOk[i])
                 PyErr_Print();        // TODO: fetch errors and reset on return
+
             countEval[i] = true;
         }
         Py_XDECREF(args);
@@ -253,87 +299,147 @@ static PyObject* minimize(PyObject* /* dummy */, PyObject* args, PyObject* kwds)
     // get initial
         double* par = nullptr;
         Py_ssize_t npar = GetBuffer(PyTuple_GET_ITEM(args, 1), 'd', sizeof(double), (void*&)par, true);
-        if (!par || npar == 0)
-            return nullptr;       // error already set
+        if (!par || npar == 0) {
+            PyErr_SetString(PyExc_TypeError, "unable to convert parameter array");
+            return nullptr;
+        }
 
         Point x0(npar);
         for (Py_ssize_t i = 0; i < npar; ++i)
             x0[i] = par[i];
         params->setAttributeValue("X0", x0);
 
-    // get the bounds buffers
+    // get the bounds buffers, if provided
         double *lower = nullptr, *upper = nullptr;
         for (Py_ssize_t iarg : {2, 3}) {
+            if (PyTuple_GET_ITEM(args, iarg) == Py_None)
+                continue;
+
             void*& bounds = (void*&)(iarg == 2 ? lower : upper);
             Py_ssize_t nbounds = GetBuffer(PyTuple_GET_ITEM(args, iarg), 'd', sizeof(double), bounds, true);
-            if (!bounds || nbounds == 0)
-                return nullptr;       // error already set
+            if (!bounds || nbounds == 0) {
+                PyErr_SetString(PyExc_TypeError, "unable to convert bounds array");
+                return nullptr;
+            }
 
             if (npar != nbounds) {
-                 PyErr_Format(PyExc_ValueError,
-                     "length of initial (" PY_SSIZE_T_FORMAT ") and %s bounds (" PY_SSIZE_T_FORMAT ") do not match",
-                     npar, (iarg == 2 ? "lower" : "upper"), nbounds);
-                 return nullptr;
+                PyErr_Format(PyExc_ValueError,
+                    "length of initial (" PY_SSIZE_T_FORMAT ") and %s bounds (" PY_SSIZE_T_FORMAT ") do not match",
+                    npar, (iarg == 2 ? "lower" : "upper"), nbounds);
+                return nullptr;
             }
         }
 
-        ArrayOfDouble lbounds(npar);
-        for (Py_ssize_t i = 0; i < npar; ++i)
-            lbounds[i] = lower[i];
-        params->setAttributeValue("LOWER_BOUND", lbounds);
+        if (lower) {
+            ArrayOfDouble lbounds(npar);
+            for (Py_ssize_t i = 0; i < npar; ++i)
+                lbounds[i] = lower[i];
+            params->setAttributeValue("LOWER_BOUND", lbounds);
+        }
 
-        ArrayOfDouble ubounds(npar);
-        for (Py_ssize_t i = 0; i < npar; ++i) {
-            ubounds[i] = upper[i];
-            if (ubounds[i] <= lbounds[i]) {
-                 PyErr_Format(PyExc_ValueError,
-                     "for element " PY_SSIZE_T_FORMAT ", upper bound(%d) is not larger than lower bound (%d)",
-                     i, ubounds[i].todouble(), lbounds[i].todouble());
-                 return nullptr;
+        if (upper) {
+            ArrayOfDouble ubounds(npar);
+            for (Py_ssize_t i = 0; i < npar; ++i) {
+                if (lower && upper[i] <= lower[i]) {
+                    PyErr_Format(PyExc_ValueError,
+                        "for element " PY_SSIZE_T_FORMAT ", upper bound (%d) is smaller than lower bound (%d)",
+                        i, upper[i], lower[i]);
+                    return nullptr;
+                }
+                ubounds[i] = upper[i];
             }
+            params->setAttributeValue("UPPER_BOUND", ubounds);
         }
-        params->setAttributeValue("UPPER_BOUND", ubounds);
 
     // process known options (allow unknown options to pass as strings)
-        bool options_ok = true;
-        PyObject* items = PyDict_Items(kwds);
-        for (Py_ssize_t i = 0; i < PyList_GET_SIZE(items); ++i) {
-            PyObject* pair = PyList_GET_ITEM(items, i);
+        std::vector<bool> options_ok{
+            true,  /* all okay */
+            false, /* have BB_OUTPUT_TYPE */
+            false, /* have DISPLAY_DEGREE */
+            false, /* have DISPLAY_ALL_EVAL */
+            false, /* have DIMENSION */
+            false  /* have SEED */ };
+        if (kwds) {
+            PyObject* items = PyDict_Items(kwds);
+            for (Py_ssize_t i = 0; i < PyList_GET_SIZE(items); ++i) {
+                PyObject* pair = PyList_GET_ITEM(items, i);
 
-            const char* key = PyCompat_PyText_AsString(PyTuple_GET_ITEM(pair, 0));
-            if (!key) {
-                options_ok = false;
-                break;
-            }
-
-            if (strcmp(key, "MAX_BB_EVAL") == 0) {
-                long budget = PyInt_AsLong(PyTuple_GET_ITEM(pair, 1));
-                if (budget == -1 && PyErr_Occurred()) {
-                    options_ok = false;
+                const char* key = PyCompat_PyText_AsString(PyTuple_GET_ITEM(pair, 0));
+                if (!key) {
+                    options_ok[0] = false;
                     break;
                 }
-                params->setAttributeValue("MAX_BB_EVAL", (int)budget);
-            } else {
-                const char* value = PyCompat_PyText_AsString(PyTuple_GET_ITEM(pair, 1));
-                if (!value) {
-                    PyErr_Format(PyExc_TypeError, "string expected for value of \'%s\'", key);
-                    options_ok = false;
-                    break;
+
+                if (strcmp(key, "MAX_BB_EVAL") == 0 || strcmp(key, "MAX_EVAL") == 0) {
+                    long budget = PyInt_AsLong(PyTuple_GET_ITEM(pair, 1));
+                    if (budget == -1 && PyErr_Occurred()) {
+                        options_ok[0] = false;
+                        break;
+                    }
+                    params->setAttributeValue("MAX_BB_EVAL", (int)budget);
+
+                // have to set MAX_EVAL along side MAX_BB_EVAL, unless overridden in options
+                    if (strcmp(key, "MAX_BB_EVAL") == 0) {
+                        PyObject* v = PyDict_GetItemString(kwds, "MAX_EVAL");
+                        if (!v) {
+                            PyErr_Clear();
+                            params->setAttributeValue("MAX_EVAL", (size_t)100*budget);
+                        }
+                    }
+                } else if (strcmp(key, "BB_OUTPUT_TYPE") == 0) {
+                    const char* value = PyCompat_PyText_AsString(PyTuple_GET_ITEM(pair, 1));
+                    if (!value) {
+                        PyErr_Format(PyExc_TypeError, "BB_OUTPUT_TYPE value field should be string");
+                        options_ok[0] = false;
+                        break;
+                    }
+                    options_ok[1] = true;
+                    params->setAttributeValue("BB_OUTPUT_TYPE", stringToBBOutputTypeList(value));
+                } else if (strcmp(key, "SEED") == 0) {
+                    const char* value = PyCompat_PyText_AsString(PyTuple_GET_ITEM(pair, 1));
+                    if (!value) {
+                        PyErr_Clear();
+                        long seed = PyInt_AsLong(PyTuple_GET_ITEM(pair, 1));
+                        if (seed == -1 && PyErr_Occurred()) {
+                             options_ok[0] = false;
+                             break;
+                        }
+                        params->setAttributeValue("SEED", (int)seed);
+                    } else
+                        params->readParamLine(std::string(key) + " " + value);
+                    options_ok[5] = true;
+                } else {
+                    if (strcmp(key, "BB_OUTPUT_TYPE") == 0)        options_ok[1] = true;
+                    else if (strcmp(key, "DISPLAY_DEGREE") == 0)   options_ok[2] = true;
+                    else if (strcmp(key, "DISPLAY_ALL_EVAL") == 0) options_ok[3] = true;
+                    else if (strcmp(key, "DIMENSION") == 0)        options_ok[4] = true;
+
+                    const char* value = PyCompat_PyText_AsString(PyTuple_GET_ITEM(pair, 1));
+                    if (!value) {
+                        PyErr_Format(PyExc_TypeError, "string expected for value of \'%s\'", key);
+                        options_ok[0] = false;
+                        break;
+                    }
+                    params->readParamLine(std::string(key) + " " + value);    // may fail on check later
                 }
-                params->readParamLine(std::string(key) + " " + value);    // may fail on check later
             }
+            Py_DECREF(items);
         }
-        Py_DECREF(items);
 
-        if (!options_ok)
-            return nullptr;
+        if (!options_ok[0])
+            return nullptr;  // error already set
 
     // add defaults/implied parameters
-        params->setAttributeValue("BB_OUTPUT_TYPE", stringToBBOutputTypeList("OBJ"));
-        params->setAttributeValue("DISPLAY_DEGREE", 0);              // pending verbose flag
-        params->setAttributeValue("DISPLAY_ALL_EVAL", false);        // id.
-        params->setAttributeValue("DIMENSION", (size_t)npar);
-        RNG::resetPrivateSeedToDefault();
+        if (!options_ok[1])
+            params->setAttributeValue("BB_OUTPUT_TYPE", stringToBBOutputTypeList("OBJ"));
+        if (!options_ok[2]) params->setAttributeValue("DISPLAY_DEGREE", 0);
+        if (!options_ok[3]) params->setAttributeValue("DISPLAY_ALL_EVAL", false);
+        if (!options_ok[4]) params->setAttributeValue("DIMENSION", (size_t)npar);
+
+        params->getPbParams()->setAttributeValue("GRANULARITY", NOMAD::ArrayOfDouble(npar, 0.0000001));
+
+        //if (!options_ok[5])
+            RNG::resetPrivateSeedToDefault();
 
     // verify (should never fail at this point)
         params->checkAndComply();
@@ -347,6 +453,10 @@ static PyObject* minimize(PyObject* /* dummy */, PyObject* args, PyObject* kwds)
         // non-existent / nothing to clear
         }
 
+    // resetting the evaluator control is necessary as MainStep will otherwise not
+    // re-initialize it, leaving the old "stop reasons" in place, failing the run
+        EvcInterface::setEvaluatorControl(nullptr);
+
     // create and configure NOMAD instance (outside of thread blocks b/c it will own a Python object)
         MainStep cnomad;
         cnomad.setAllParameters(params);
@@ -354,17 +464,18 @@ static PyObject* minimize(PyObject* /* dummy */, PyObject* args, PyObject* kwds)
             std::make_unique<PyCallback>(params->getEvalParams(), PyTuple_GET_ITEM(args, 0)));
 
         bool run_ok = false;
-        Py_BEGIN_ALLOW_THREADS
+        {
+        PyUnGILRAII m;
+        EvcInterfaceReset e;
 
         cnomad.start();
         run_ok = cnomad.run();
         cnomad.end();
-
-        Py_END_ALLOW_THREADS
+        }
 
         if (run_ok) {
             std::vector<EvalPoint> evpl;
-            auto nf = CacheBase::getInstance()->findBestFeas(evpl, Point(), EvalType::BB);
+            auto nf = CacheBase::getInstance()->findBestFeas(evpl, Point(), EvalType::BB, nullptr);
             if (0 < nf) {
                 EvalPoint evres = evpl[0];
                 result = PyTuple_New(2);
@@ -384,14 +495,19 @@ static PyObject* minimize(PyObject* /* dummy */, PyObject* args, PyObject* kwds)
                 PyErr_SetString(PyExc_RuntimeError, "no feasible point");
         }
 
+    } catch (const NOMAD::Exception& e) {
+        PyErr_Format(PyExc_RuntimeError, "%s", e.what());
+        return nullptr;
     } catch (const std::exception& e) {
-        PyErr_Format(PyExc_TypeError, "C++ exception: %s", e.what());
+        PyErr_Format(PyExc_RuntimeError, "[C++] %s", e.what());
         return nullptr;
     } catch (...) {
-        PyErr_SetString(PyExc_TypeError, "Unknown C++ exception");
+        PyErr_SetString(PyExc_RuntimeError, "Unknown C++ exception");
         return nullptr;
     }
 
+    if (!result && !PyErr_Occurred())
+        PyErr_SetString(PyExc_RuntimeError, "NOMAD run failed");
     return result;
 }
 
@@ -448,7 +564,18 @@ extern "C" void initlibsqnomad()
     using namespace SQNomad;
 
 // setup interpreter
-    PyEval_InitThreads();
+    if (!Py_IsInitialized()) {
+    // this happens if Cling comes in first
+#if PY_VERSION_HEX < 0x03020000
+        PyEval_InitThreads();
+#endif
+        Py_Initialize();
+#if PY_VERSION_HEX >= 0x03020000
+#if PY_VERSION_HEX < 0x03090000
+        PyEval_InitThreads();
+#endif
+#endif
+    }
 
 // force numpy
     import_array()
